// A library to parallelize code on Tilera TILE-Gx easily.
// Written by Nils Liaaen Corneliusen 2014.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// Please refer to the article at https://www.ignorantus.com/pages/tilegx_bilinear/ for more information.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>
#include <pthread.h>
#include <tmc/cpus.h>
#include <tmc/task.h>
#include <tmc/udn.h>

#include "par.h"

unsigned int par_cores_total  = 0;
unsigned int par_cores_max    = 0;
unsigned int par_cores_active = 0;

void par_sendjob( void *func, void *data, unsigned int len )
{
    int core;
    DynamicHeader header;

    assert( len <= DATA_LEN_MAX );

    // still available cores?
    if( par_cores_active < par_cores_max ) {
        // yes: start next
        par_cores_active++;
        core = par_cores_active;
    } else {
        // no: wait for a core to be done
        core = tmc_udn0_receive();
    }

    header = tmc_udn_header_from_cpu( core );

    tmc_udn_send_2( header, UDN0_DEMUX_TAG, (uint64_t)func, len );
    tmc_udn_send_buffer( header, UDN0_DEMUX_TAG, data, (len+7)/8 );
}

void par_wait( void )
{
    for( unsigned int i = 0; i < par_cores_active; i++ )
        tmc_udn0_receive();

    par_cores_active = 0;
}

void par_set_cores( unsigned int cores )
{
    par_cores_max = cores > par_cores_total ? par_cores_total : cores;
}

static void *par_thread( void *arg )
{
    int coreid = (int)(uint64_t)arg;

    if( tmc_cpus_set_my_cpu( coreid ) != 0 )
        tmc_task_die( "Core %d: tmc_cpus_set_my_cpu() failed.", coreid );

    if( tmc_udn_activate() < 0 )
        tmc_task_die( "Core %d: tmc_udn_activate() failed.", coreid );

    // initialize stuff
    DynamicHeader header = tmc_udn_header_from_cpu( 0 );
    void *data = malloc( DATA_LEN_MAX );
    if( data == NULL )
        tmc_task_die( "Core %d: malloc() failed.", coreid );

    // ready to start
    tmc_udn_send_1( header, UDN0_DEMUX_TAG, coreid );

    while( true ) {
        void (* func)(void *data, int len);
        unsigned int len;

        // get job data
        func = (void *)tmc_udn0_receive();
        if( func == (void *)0xffff ) exit( 0 );

        len = tmc_udn0_receive();
        assert( len <= DATA_LEN_MAX );
        tmc_udn0_receive_buffer( data, (len+7)/8 );

        // start job
        func( data, len );

        // say we're done
        tmc_udn_send_1( header, UDN0_DEMUX_TAG, coreid );

    }

    return NULL;
}

int par_init( unsigned int cores )
{
    int rc;
    cpu_set_t cpus;

    par_cores_total = cores;
    par_cores_max   = cores;

    if( tmc_cpus_get_my_affinity( &cpus ) != 0 ) {
        printf( "Host: tmc_cpus_get_my_affinity() failed.\n" );
        return 1;
    }

    // Add 1 for host
    if( tmc_cpus_count( &cpus ) < cores+1 ) {
        printf( "Host: Insufficient cpus (%d < %d).\n", tmc_cpus_count(&cpus), cores );
        return 1;
    }

    if( tmc_udn_init( &cpus ) < 0 ) {
        printf( "Host: tmc_udn_init() failed.\n" );
        return 1;
    }

    pthread_attr_t attr;
    rc = pthread_attr_init( &attr );
    if( rc != 0 ) {
        perror( "pthread_attr_init" );
        return rc;
    }

    // Threads are assigned from core 1 upwards
    for( uint64_t rank = 1; rank <= cores; rank++ ) {
        pthread_t thread_id;
        rc = pthread_create( &thread_id, &attr, &par_thread, (void *)rank );
        if( rc != 0 ) {
            perror( "pthread_create" );
            pthread_attr_destroy( &attr );
            return rc;
        }
    }

    rc = pthread_attr_destroy( &attr );

    // Put host on core 0
    if( tmc_cpus_set_my_cpu( 0 ) != 0 ) {
        printf( "Host: tmc_cpus_set_my_cpu() failed.\n" );
        return 1;
    }

    if( tmc_udn_activate() < 0 ) {
        printf( "Host: tmc_udn_activate() failed.\n" );
        return 1;
    }

    // wait for initialization done
    for( unsigned int i = 0; i < cores; i++ )
        tmc_udn0_receive();

    return 0;
}

void par_shutdown( void )
{
    par_wait();
    for( unsigned int core = 1; core < par_cores_max+1; core++ ) {
        DynamicHeader header = tmc_udn_header_from_cpu( core );
        tmc_udn_send_1( header, UDN0_DEMUX_TAG, 0xffff );
    }
}
