// A bilinear scaler for Tilera TILE-Gx.
// Written by Nils Liaaen Corneliusen 2014.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// Please refer to the article at https://www.ignorantus.com/pages/tilegx_bilinear/ for more information.

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <alloca.h>
#include <arch/spr.h>
#include <postinc.h>

#include "scaler.h"
#include "par.h"


void scale_bilinear( bmp_argb *src, bmp_argb *dst, unsigned int startrow, unsigned int height )
{
    uint64_t roundval = 0x0080008000800080;

    int xadd = (src->w<<16)/dst->w;
    int yadd = (src->h<<16)/dst->h;

    int xoffset = (xadd>>1)-(1<<15);
    int yoffset = (yadd>>1)-(1<<15);

    int dstw  = dst->w;

    for( int y = startrow; y < startrow + height; y += 2 ) {

        int yy0 = yoffset+ y   *yadd;
        int yy1 = yoffset+(y+1)*yadd;
        int ypos0 = yy0>>16; if( ypos0 < 0 ) ypos0 = 0;
        int ypos1 = yy1>>16; if( ypos1 < 0 ) ypos1 = 0;

        // find rows to fetch data from
        uint8_t *row00 = src->argb + src->stride*(ypos0+0);
        uint8_t *row01 = src->argb + src->stride*(ypos0+1);
        uint8_t *row10 = src->argb + src->stride*(ypos1+0);
        uint8_t *row11 = src->argb + src->stride*(ypos1+1);

        // get vertical weights
        uint64_t vmul0, vmul1, vmul2, vmul3;
        vmul0 = (yy0>>8)&0xff;
        vmul0 = __insn_shufflebytes( vmul0, 0, 0 );
        vmul1 = 0xffffffff - vmul0;
        vmul2 = (yy1>>8)&0xff;
        vmul2 = __insn_shufflebytes( vmul2, 0, 0 );
        vmul3 = 0xffffffff - vmul2;

        uint8_t *dp0 = dst->argb +  y   *dst->stride;
        uint8_t *dp1 = dst->argb + (y+1)*dst->stride;

        for( int x = 0; x < dstw; x += 2 ) {
            uint8_t *tmp;
            uint64_t p00, p01, p02, p03;
            uint64_t p10, p11, p12, p13;
            uint64_t p20, p21, p22, p23;
            uint64_t p30, p31, p32, p33;
            uint64_t v00, v01;
            uint64_t r00, r01, r10, r11;
            uint64_t hmul0,hmul1,hmul2,hmul3;

            int xx0 = xoffset+ x   *xadd;
            int xx1 = xoffset+(x+1)*xadd;
            int xpos0 = (xx0>>16)*4; if( xpos0 < 0 ) xpos0 = 0;
            int xpos1 = (xx1>>16)*4; if( xpos1 < 0 ) xpos1 = 0;

            // load (2*2)*(2*2) pixels so it can be unrolled a bit
            tmp = row00 + xpos0; p00 = __insn_ld4u_add( tmp, 4 ); p01 = __insn_ld4u( tmp );
            tmp = row00 + xpos1; p02 = __insn_ld4u_add( tmp, 4 ); p03 = __insn_ld4u( tmp );
            tmp = row01 + xpos0; p10 = __insn_ld4u_add( tmp, 4 ); p11 = __insn_ld4u( tmp );
            tmp = row01 + xpos1; p12 = __insn_ld4u_add( tmp, 4 ); p13 = __insn_ld4u( tmp );
            tmp = row10 + xpos0; p20 = __insn_ld4u_add( tmp, 4 ); p21 = __insn_ld4u( tmp );
            tmp = row10 + xpos1; p22 = __insn_ld4u_add( tmp, 4 ); p23 = __insn_ld4u( tmp );
            tmp = row11 + xpos0; p30 = __insn_ld4u_add( tmp, 4 ); p31 = __insn_ld4u( tmp );
            tmp = row11 + xpos1; p32 = __insn_ld4u_add( tmp, 4 ); p33 = __insn_ld4u( tmp );

            // get horizontal weights
            hmul0 = (((uint64_t)xx0)>>8)&0xff;
            hmul0 = __insn_shufflebytes( hmul0, 0, 0 );
            hmul1 = 0xffffffff - hmul0;
            hmul2 = (((uint64_t)xx1)>>8)&0xff;
            hmul2 = __insn_shufflebytes( hmul2, 0, 0 );
            hmul3 = 0xffffffff - hmul2;

            // calculate 2*2 output pixels
            v00 = __insn_v1mulu( p00, vmul1 ) + __insn_v1mulu( p10, vmul0 );
            v01 = __insn_v1mulu( p01, vmul1 ) + __insn_v1mulu( p11, vmul0 );
            v00 =  __insn_v2packh( 0, v00 + __insn_v2shrui( v00, 8 ) + roundval );
            v01 =  __insn_v2packh( 0, v01 + __insn_v2shrui( v01, 8 ) + roundval );

            r00 = __insn_v1mulu( v00, hmul1 ) + __insn_v1mulu( v01, hmul0 );
            r00 = r00 + __insn_v2shrui( r00, 8 ) + roundval;

            v00 = __insn_v1mulu( p02, vmul1 ) + __insn_v1mulu( p12, vmul0 );
            v01 = __insn_v1mulu( p03, vmul1 ) + __insn_v1mulu( p13, vmul0 );
            v00 =  __insn_v2packh( 0, v00 + __insn_v2shrui( v00, 8 ) + roundval );
            v01 =  __insn_v2packh( 0, v01 + __insn_v2shrui( v01, 8 ) + roundval );

            r01 = __insn_v1mulu( v00, hmul3 ) + __insn_v1mulu( v01, hmul2 );
            r01 = r01 + __insn_v2shrui( r01, 8 ) + roundval;

            v00 = __insn_v1mulu( p20, vmul3 ) + __insn_v1mulu( p30, vmul2 );
            v01 = __insn_v1mulu( p21, vmul3 ) + __insn_v1mulu( p31, vmul2 );
            v00 =  __insn_v2packh( 0, v00 + __insn_v2shrui( v00, 8 ) + roundval );
            v01 =  __insn_v2packh( 0, v01 + __insn_v2shrui( v01, 8 ) + roundval );

            r10 = __insn_v1mulu( v00, hmul1 ) + __insn_v1mulu( v01, hmul0 );
            r10 = r10 + __insn_v2shrui( r10, 8 ) + roundval;

            v00 = __insn_v1mulu( p22, vmul3 ) + __insn_v1mulu( p32, vmul2 );
            v01 = __insn_v1mulu( p23, vmul3 ) + __insn_v1mulu( p33, vmul2 );
            v00 =  __insn_v2packh( 0, v00 + __insn_v2shrui( v00, 8 ) + roundval );
            v01 =  __insn_v2packh( 0, v01 + __insn_v2shrui( v01, 8 ) + roundval );

            r11 = __insn_v1mulu( v00, hmul3 ) + __insn_v1mulu( v01, hmul2 );
            r11 = r11 + __insn_v2shrui( r11, 8 ) + roundval;

            // pack and store
            __insn_st_add( dp0, __insn_v2packh( r01, r00 ), 8 );
            __insn_st_add( dp1, __insn_v2packh( r11, r10 ), 8 );

        }

    }

}


typedef struct {
    bmp_argb src;
    bmp_argb dst;
    unsigned int startrow;
    unsigned int height;
} sc_info;

static void scale_bilinear_thread( void *data, unsigned int len )
{
    sc_info *sc = data;

    scale_bilinear( &sc->src, &sc->dst, sc->startrow, sc->height );
}

void scale_bilinear_parallel( bmp_argb *src, bmp_argb *dst, unsigned int lines )
{
    sc_info sc;
    sc.src = *src;
    sc.dst = *dst;

    for( unsigned int startrow = 0; startrow < dst->h; startrow += lines ) {

        sc.startrow = startrow;
        sc.height   = dst->h - startrow < lines ? dst->h - startrow : lines;

        par_sendjob( scale_bilinear_thread, &sc, sizeof(sc) );
    }

    par_wait();
}



static inline void scale_bilinear_vertical( bmp_argb *src, bmp_argb *dst, uint8_t *out0, unsigned int startrow )
{
    uint64_t roundval = 0x0080008000800080;
    int yadd = (src->h<<16)/dst->h;
    int yoffset = (yadd>>1)-(1<<15);
    int srcw  = src->w;

    int yy0 = yoffset+ startrow   *yadd;
    int yy1 = yoffset+(startrow+1)*yadd;
    int ypos0 = yy0>>16; if( ypos0 < 0 ) ypos0 = 0;
    int ypos1 = yy1>>16; if( ypos1 < 0 ) ypos1 = 0;

    // find rows to fetch data from
    uint8_t *row00 = src->argb + src->stride*(ypos0+0);
    uint8_t *row01 = src->argb + src->stride*(ypos0+1);
    uint8_t *row10 = src->argb + src->stride*(ypos1+0);
    uint8_t *row11 = src->argb + src->stride*(ypos1+1);

    // get vertical weights
    uint64_t vmul0, vmul1, vmul2, vmul3;
    vmul0 = __insn_shufflebytes( (yy0>>8)&0xff, 0, 0 ); vmul1 = 0xffffffff - vmul0;
    vmul2 = __insn_shufflebytes( (yy1>>8)&0xff, 0, 0 ); vmul3 = 0xffffffff - vmul2;

    uint8_t *out1 = out0 + src->stride;

    for( int x = 0; x < srcw; x += 4 ) {

        uint64_t p001 = __insn_ld_add( row00, 8 ); uint64_t p023 = __insn_ld_add( row00, 8 );
        uint64_t p101 = __insn_ld_add( row01, 8 ); uint64_t p123 = __insn_ld_add( row01, 8 );
        uint64_t p201 = __insn_ld_add( row10, 8 ); uint64_t p223 = __insn_ld_add( row10, 8 );
        uint64_t p301 = __insn_ld_add( row11, 8 ); uint64_t p323 = __insn_ld_add( row11, 8 );

        uint64_t v001l = __insn_v1mulu( p001,     vmul1 ) + __insn_v1mulu( p101,     vmul0 );
        uint64_t v001h = __insn_v1mulu( p001>>32, vmul1 ) + __insn_v1mulu( p101>>32, vmul0 );
        uint64_t v023l = __insn_v1mulu( p023,     vmul1 ) + __insn_v1mulu( p123,     vmul0 );
        uint64_t v023h = __insn_v1mulu( p023>>32, vmul1 ) + __insn_v1mulu( p123>>32, vmul0 );
        uint64_t v101l = __insn_v1mulu( p201,     vmul3 ) + __insn_v1mulu( p301,     vmul2 );
        uint64_t v101h = __insn_v1mulu( p201>>32, vmul3 ) + __insn_v1mulu( p301>>32, vmul2 );
        uint64_t v123l = __insn_v1mulu( p223,     vmul3 ) + __insn_v1mulu( p323,     vmul2 );
        uint64_t v123h = __insn_v1mulu( p223>>32, vmul3 ) + __insn_v1mulu( p323>>32, vmul2 );

        uint64_t r001l =  v001l + __insn_v2shrui( v001l, 8 ) + roundval;
        uint64_t r001h =  v001h + __insn_v2shrui( v001h, 8 ) + roundval;
        uint64_t r023l =  v023l + __insn_v2shrui( v023l, 8 ) + roundval;
        uint64_t r023h =  v023h + __insn_v2shrui( v023h, 8 ) + roundval;
        uint64_t r101l =  v101l + __insn_v2shrui( v101l, 8 ) + roundval;
        uint64_t r101h =  v101h + __insn_v2shrui( v101h, 8 ) + roundval;
        uint64_t r123l =  v123l + __insn_v2shrui( v123l, 8 ) + roundval;
        uint64_t r123h =  v123h + __insn_v2shrui( v123h, 8 ) + roundval;

        __insn_st_add( out0, __insn_v2packh( r001h, r001l ), 8 );
        __insn_st_add( out0, __insn_v2packh( r023h, r023l ), 8 );
        __insn_st_add( out1, __insn_v2packh( r101h, r101l ), 8 );
        __insn_st_add( out1, __insn_v2packh( r123h, r123l ), 8 );

    }
}


static inline void scale_bilinear_horizontal( bmp_argb *src, bmp_argb *dst, uint8_t *in0, unsigned int startrow )
{
    uint64_t roundval = 0x0080008000800080;
    int xadd = (src->w<<16)/dst->w;
    int xoffset = (xadd>>1)-(1<<15);
    int dstw = dst->w;

    uint8_t *row00 = in0;
    uint8_t *row01 = in0 + src->stride;

    uint8_t *out0 = dst->argb +  startrow   * dst->stride;
    uint8_t *out1 = dst->argb + (startrow+1)* dst->stride;

    for( int x = 0; x < dstw; x += 4 ) {
        uint8_t *tmp;
        uint64_t p00,p01,p02,p03,p04,p05,p06,p07;
        uint64_t p10,p11,p12,p13,p14,p15,p16,p17;
        uint64_t hmul0,hmul1,hmul2,hmul3,hmul4,hmul5,hmul6,hmul7;

        int xx0 = xoffset+ x   *xadd;
        int xx1 = xoffset+(x+1)*xadd;
        int xx2 = xoffset+(x+2)*xadd;
        int xx3 = xoffset+(x+3)*xadd;
        int xpos0 = (xx0>>16)*4; if( xpos0 < 0 ) xpos0 = 0;
        int xpos1 = (xx1>>16)*4; if( xpos1 < 0 ) xpos1 = 0;
        int xpos2 = (xx2>>16)*4; if( xpos2 < 0 ) xpos2 = 0;
        int xpos3 = (xx3>>16)*4; if( xpos3 < 0 ) xpos3 = 0;

        tmp = row00 + xpos0; p00 = __insn_ld4u_add( tmp, 4 ); p01 = __insn_ld4u( tmp );
        tmp = row00 + xpos1; p02 = __insn_ld4u_add( tmp, 4 ); p03 = __insn_ld4u( tmp );
        tmp = row00 + xpos2; p04 = __insn_ld4u_add( tmp, 4 ); p05 = __insn_ld4u( tmp );
        tmp = row00 + xpos3; p06 = __insn_ld4u_add( tmp, 4 ); p07 = __insn_ld4u( tmp );
        tmp = row01 + xpos0; p10 = __insn_ld4u_add( tmp, 4 ); p11 = __insn_ld4u( tmp );
        tmp = row01 + xpos1; p12 = __insn_ld4u_add( tmp, 4 ); p13 = __insn_ld4u( tmp );
        tmp = row01 + xpos2; p14 = __insn_ld4u_add( tmp, 4 ); p15 = __insn_ld4u( tmp );
        tmp = row01 + xpos3; p16 = __insn_ld4u_add( tmp, 4 ); p17 = __insn_ld4u( tmp );

        hmul0 = __insn_shufflebytes( (((uint64_t)xx0)>>8)&0xff, 0, 0 ); hmul1 = 0xffffffff - hmul0;
        hmul2 = __insn_shufflebytes( (((uint64_t)xx1)>>8)&0xff, 0, 0 ); hmul3 = 0xffffffff - hmul2;
        hmul4 = __insn_shufflebytes( (((uint64_t)xx2)>>8)&0xff, 0, 0 ); hmul5 = 0xffffffff - hmul4;
        hmul6 = __insn_shufflebytes( (((uint64_t)xx3)>>8)&0xff, 0, 0 ); hmul7 = 0xffffffff - hmul6;

        uint64_t r00 = __insn_v1mulu( p00, hmul1 ) + __insn_v1mulu( p01, hmul0 );
        uint64_t r01 = __insn_v1mulu( p02, hmul3 ) + __insn_v1mulu( p03, hmul2 );
        uint64_t r02 = __insn_v1mulu( p04, hmul5 ) + __insn_v1mulu( p05, hmul4 );
        uint64_t r03 = __insn_v1mulu( p06, hmul7 ) + __insn_v1mulu( p07, hmul6 );
        uint64_t r10 = __insn_v1mulu( p10, hmul1 ) + __insn_v1mulu( p11, hmul0 );
        uint64_t r11 = __insn_v1mulu( p12, hmul3 ) + __insn_v1mulu( p13, hmul2 );
        uint64_t r12 = __insn_v1mulu( p14, hmul5 ) + __insn_v1mulu( p15, hmul4 );
        uint64_t r13 = __insn_v1mulu( p16, hmul7 ) + __insn_v1mulu( p17, hmul6 );

        r00 = r00 + __insn_v2shrui( r00, 8 ) + roundval;
        r01 = r01 + __insn_v2shrui( r01, 8 ) + roundval;
        r02 = r02 + __insn_v2shrui( r02, 8 ) + roundval;
        r03 = r03 + __insn_v2shrui( r03, 8 ) + roundval;
        r10 = r10 + __insn_v2shrui( r10, 8 ) + roundval;
        r11 = r11 + __insn_v2shrui( r11, 8 ) + roundval;
        r12 = r12 + __insn_v2shrui( r12, 8 ) + roundval;
        r13 = r13 + __insn_v2shrui( r13, 8 ) + roundval;

        __insn_st_add( out0, __insn_v2packh( r01, r00 ), 8 );
        __insn_st_add( out0, __insn_v2packh( r03, r02 ), 8 );
        __insn_st_add( out1, __insn_v2packh( r11, r10 ), 8 );
        __insn_st_add( out1, __insn_v2packh( r13, r12 ), 8 );
    }

}


static void scale_bilinear_2pass( bmp_argb *src, bmp_argb *dst, unsigned int startrow, unsigned int height )
{
    uint8_t *vbuf = alloca( src->stride*2 );

    for( int y = startrow; y < startrow + height; y += 2 ) {

        scale_bilinear_vertical( src, dst, vbuf, y );

        scale_bilinear_horizontal( src, dst, vbuf, y );

    }

}


static void scale_bilinear_thread_2pass( void *data, unsigned int len )
{
    sc_info *sc = data;

    scale_bilinear_2pass( &sc->src, &sc->dst, sc->startrow, sc->height );
}


void scale_bilinear_parallel_2pass( bmp_argb *src, bmp_argb *dst, unsigned int lines )
{
    sc_info sc;
    sc.src = *src;
    sc.dst = *dst;

    for( unsigned int startrow = 0; startrow < dst->h; startrow += lines ) {

        sc.startrow = startrow;
        sc.height   = dst->h - startrow < lines ? dst->h - startrow : lines;

        par_sendjob( scale_bilinear_thread_2pass, &sc, sizeof(sc) );
    }

    par_wait();
}

